#!/usr/bin/perl

use strict;

# Call this as:  ./do_test.pl eth1 eth3

my $dev1 = "$ARGV[0]";
my $dev2 = "$ARGV[1]";
my $test_case = 0;
my $duration = 20 * 1000; # 20 seconds.

if (@ARGV > 2) {
  $test_case = "$ARGV[2]";
}

if (@ARGV > 3) {
  $duration = $ARGV[3];
}

if (@ARGV < 2) {
  print "Usage: ./to_test.pl eth1 eth3 [test-case-number]\n";
  print "  Test-cases:\n";
  print "  0:  Ensure receipt of tags when bridging ether devices\n";
  print "      with an un-used VLAN on one device.  Uses recvmsg (aux-data)\n";
  print "  1:  Ensure receipt of tags when bridging ether devices\n";
  print "      with an un-used VLAN on one device.  Uses recvfrom (no aux-data)\n";
  print "  2:  Ensure receipt of tags when bridging ether devices\n";
  print "      without any VLANs configured.  Uses recvmsg (aux-data)\n";
  print "  3:  Ensure receipt of tags when bridging ether devices\n";
  print "      without any VLANs configured.  Uses recvfrom (no aux-data)\n";
  print "  4:  Ensure receipt of tags when bridging ether devices\n";
  print "      with bridged VLAN on one device.  Uses recvmsg (aux-data)\n";
  print "  5:  Ensure receipt of tags when bridging ether devices\n";
  print "      with an bridged VLAN on one device.  Uses recvfrom (no aux-data)\n";
  exit(2);
}

print "Test info:\n";
print `date`;
print `uname -a`;
print "\nInterface A: $dev1\n";
print `ethtool -i $dev1`;
print "\nInterface B: $dev2:\n";
print `ethtool -i $dev2`;
print "\n";

print "Bringing up interfaces, making them PROMISC.\n";
system("ifconfig $dev1 up");
system("ifconfig $dev2 up");
system("ifconfig $dev1 promisc");
system("ifconfig $dev2 promisc");

my $extra = "";
if (($test_case == 1) || ($test_case == 3) || ($test_case == 5)) {
  # recvfrom
  $extra = "-r";
}

# Cleanup any old VLANs lying around.
system("ip link del $dev1.9");
system("ip link del $dev1.7");


if (($test_case == 0) || ($test_case == 1)) {
  # Test with VLAN existing, but VLAN is not used by generator
  print "Adding VLAN 9 to $dev1 to possibly enable tag stripping in driver.\n";
  system("ip link add link $dev1 up name $dev1.9 type vlan id 9");
  system("./ubridge -1 $dev1 -2 $dev2 --duration $duration $extra");
}
elsif (($test_case == 2) || ($test_case == 3)) {
  # Test with VLAN existing
  system("./ubridge -1 $dev1 -2 $dev2 --duration $duration $extra");
}
elsif (($test_case == 4) || ($test_case == 5)) {
  # Test with VLAN existing, traffic is on VLAN 7
  print "Adding VLAN 7 to $dev1 to possibly enable tag stripping in driver.\n";
  system("ip link add link $dev1 up name $dev1.7 type vlan id 7");
  system("./ubridge -1 $dev1 -2 $dev2 --duration $duration $extra");
}
else {
  print "Un-known test case: $test_case.\n";
}

print `date`;

exit(0);







