#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <net/if.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <signal.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>

#include <string>
#include <iostream>

using namespace std;

typedef unsigned long long uint64;
typedef long long int64;

// Global configurables.
string dev1 = "bad1";
string dev2 = "bad2";
bool should_use_recvfrom = false; // just reads pkt
bool should_use_recvmsg = true; // lets us get aux data
uint64 duration = 0; // forever

// Stats
uint64 rxa = 0;
uint64 rxb = 0;
uint64 rxa_vlan_hdr_fixups = 0;
uint64 rxb_vlan_hdr_fixups = 0;
uint64 txa = 0;
uint64 txb = 0;
uint64 txa_failed = 0;
uint64 txb_failed = 0;
int fd1 = -1;
int fd2 = -1;

int do_shutdown = false;
#define RCV_MSG_SIZE 64 * 1024
char rbuf[RCV_MSG_SIZE];

struct vlan_ether_header {
   char dst[6];
   char src[6];
   short type;
   short tci_vid;
   short encapsulated_type;
} __attribute__ ((__packed__));

/** Convert from mili-seconds to struct timeval. */
struct timeval ms_to_tv(uint64 ms) {
   struct timeval retval;
   uint64 tmp = ms / 1000;
   if (sizeof(retval.tv_sec) == 4) {
      // Deal somewhat with overflow
      if (tmp > 0xFFFFFFFF) {
         tmp = 0xFFFFFFFF;
      }
   }
   retval.tv_usec = (ms % 1000) * 1000;
   retval.tv_sec = tmp;
   return retval;
}

uint64 tv_to_ms(const struct timeval& tv) {
   uint64 ms = tv.tv_usec / 1000;
   ms += (uint64)tv.tv_sec * (uint64)1000;
   return ms;
}

uint64 getCurMs() {
   struct timeval tv;
   gettimeofday(&tv, NULL);
   return tv_to_ms(tv);
}

int64 parseInt64(const char* s) {
   //should parse HEX, Octal, and Decimal.  If not decimal, must start with 0x
   // Use long-long because then we can get a real value for a uint32, even if
   // it is negative.  Code using this can cast it back to un-signed int if
   // that makes sense.
   long long rv = strtoll(s, NULL, 0);
   return rv; 
}

void hexDump(unsigned char* b, int len) {
   for (int i = 0; i<len; i++) {
      printf("%02hx ", b[i]);
      if ((i & 0xF) == 0xF) {
         printf("\n");
      }
   }
   cout << endl;
}


void nonblock(int s, const char* msg) {
   int  flags;

   flags = fcntl(s, F_GETFL);
   flags |= (O_NONBLOCK | FD_CLOEXEC);
   if (fcntl(s, F_SETFL, flags) < 0) {
      cerr << "ERROR:  fcntl, executing nonblock/close_exec:  " << strerror(errno)
           << endl;
   }
}//nonblock

bool lastWasEAGAIN() {
   if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR)) {
      return true;
   }
   return false;
}

int sendPkt(int sd, const char* msg, int msg_len) {
   int r = sendto(sd, msg, msg_len, 0, NULL, 0);
   if (r < 0) {
      cerr << "Failed sending packet to socket: " << sd << " "
           << strerror(errno) << " msg-len: " << msg_len << endl;
      if (sd == fd1) {
         txa_failed++;
      }
      else {
         txb_failed++;
      }
   }
   else {
      if (sd == fd1) {
         txa++;
      }
      else {
         txb++;
      }
   }
   return r;
}

int slurpDevSocket(int sd, int peer_sd) {
   int r = 0;
   int cnt = 0;
   int rcvd = 0;
   char* buf = rbuf;

   while (cnt < 25) {
      cnt++;
      if (should_use_recvfrom) {
         struct sockaddr_ll from;
         socklen_t fromlen;
         fromlen = sizeof(from);
      
         memset(&from, 0, fromlen);
         r = recvfrom(sd, buf, RCV_MSG_SIZE, 0, (struct sockaddr*)(&from), &fromlen);
      }
      else {
         struct msghdr hdr;
         struct iovec iov;
         struct tpacket_auxdata* aux;
         char abuf[64];

         buf += 16;
         iov.iov_base = buf;
         iov.iov_len = RCV_MSG_SIZE - 16;
         memset(&hdr, 0, sizeof(hdr));
         hdr.msg_iov = &iov;
         hdr.msg_iovlen = 1;
         hdr.msg_control = abuf;
         hdr.msg_controllen = sizeof(abuf);
         r = recvmsg(sd, &hdr, MSG_DONTWAIT);
         if (r > 0) {
            // Check for vlan flag in aux data
            struct cmsghdr *cm = CMSG_FIRSTHDR(&hdr);
            while (cm) {
               struct tpacket_auxdata* aux;
               //cout << "Found cm, type: " << cm->cmsg_type << " aux: " << PACKET_AUXDATA
               //     << " len: " << cm->cmsg_len << " sizeof aux: " << sizeof(*aux) 
               //     << " sizeof cm: " << sizeof(*cm) << endl;
               if ((cm->cmsg_type == PACKET_AUXDATA) &&
                   (cm->cmsg_len >= (sizeof(*aux) + sizeof(*cm)))) {
                  aux = (struct tpacket_auxdata*)(CMSG_DATA(cm));
                  if (aux->tp_vlan_tci != 0) {
                     // re-insert the hdr for xmit
                     memmove(buf - 4, buf, 12);
                     struct vlan_ether_header* ve = (struct vlan_ether_header*)(buf - 4);
                     ve->type = htons(0x8100);
                     ve->tci_vid = htons(aux->tp_vlan_tci);
                     buf = buf - 4;
                     r += 4; //count VLAN hdr bytes
                     //cout << " aux->status: " << hex << aux->tp_status << dec << " len: " << aux->tp_len
                     //     << " snaplen: " << aux->tp_snaplen << " mac: " << aux->tp_mac << " net: "
                     //     << aux->tp_net << " tci: " << hex <<  aux->tp_vlan_tci << dec << endl;
                     //hexDump((unsigned char*)(buf), 32);
                     if (sd == fd1) {
                        rxa_vlan_hdr_fixups++;
                     }
                     else {
                        rxb_vlan_hdr_fixups++;
                     }
                  }
                  break;
               }
               cm = CMSG_NXTHDR(&hdr, cm);
            }
         }
      }            
      if (r <= 0) {
         if (!lastWasEAGAIN()) {
            cerr << "ERROR: recvfrom: " << strerror(errno) << endl;
            break;
         }
         else {
            return rcvd;
         }
      }//if an error
      else {
         rcvd++;
         // Send pkt to peer
         // TODO:  Add verification checks here to tell what kind of VLAN it is, etc?
         sendPkt(peer_sd, buf, r);
      }//else
   }//while
   return cnt;
}//slurpDevSocket

int createPacketSocket(const char* dev_name, int ether_type, int dev_idx, bool should_bind) {

   int s = socket(PF_PACKET, SOCK_RAW, htons(ether_type));
   int r; //retval

   if (s < 0) {
      cerr << "ERROR: socket:  " << strerror(errno) << endl;
      return s;
   }

   if (should_bind) {
      struct sockaddr_ll myaddr;

      memset(&myaddr, '\0', sizeof(myaddr));
      myaddr.sll_family = AF_PACKET;
      myaddr.sll_protocol = htons(ether_type);
      myaddr.sll_ifindex = dev_idx;
      
      r = bind(s, (struct sockaddr*)(&myaddr), sizeof(myaddr));
      if (r < 0) {
         cerr << "ERROR: bind:  " << strerror(errno) << endl;
         return r;
      }
   }

   if (should_use_recvmsg) {
      int opt = 1;
      if (setsockopt(s, SOL_PACKET, PACKET_AUXDATA, (char*)(&opt),
                     sizeof(opt)) < 0) {
         cerr << "ERROR:  setsockopt AUXDATA:  " << strerror(errno) << endl;
         close(s);
         return -1;
      }
   }

   nonblock(s, "create pkt socket");
   return s;
}

void printHelp() {
   cout << "ubridge:\n"
      "-1 --dev1     Side A network device.\n"
      "-2 --dev2     Side B network device.\n"
      "-d --duration How long, in miliseconds, to run before exiting.\n"
      "-h --help     Print this message.\n"
      "-x --recvmsg  Use recvmsg to read packets (default)\n"
      "              This allows getting aux packet data.\n"
      "-r --recvfrom Use recvfrom to read packets.\n"
      "\n\n"
      "NOTE:  Your interfaces must be set to PROMISC mode to bridge properly."
      "\n";
}
   
void process_args(int argc, char** argv) {
   char c;
   while (true) {
      int option_index = 0;
      static struct option long_options[] =
      {
         {"dev1", 1, 0, 0},
         {"dev2", 1, 0, 0},
         {"duration", 1, 0, 0},
         {"recvmsg", 0, 0, 0}, // x
         {"recvfrom", 0, 0, 0}, // r
         {"help", 0, 0, 0},
         {0, 0, 0, 0}
      };

      c = getopt_long (argc, argv, "1:2:d:hrx",
                       long_options, &option_index);
      if (c == -1)
         break;
      
      if (c == 0) {
         // Then we were passed a long argument (ie --foo).  So, we'll just
         // grab the first character.  Have to make sure that is unique...
         c = long_options[option_index].name[0]; //grab first char

         // Have to do special things for cmds that don't have unique
         // first letters.
         if (strcmp(long_options[option_index].name, "dev1") == 0) {
            c = '1';
         }
         else if (strcmp(long_options[option_index].name, "dev2") == 0) {
            c = '2';
         }
         else if (strcmp(long_options[option_index].name, "recvmsg") == 0) {
            c = 'x';
         }
      }

      switch (c) {
         case '1':
            dev1 = optarg;
            break;

         case '2':
            dev2 = optarg;
            break;

         case 'd':
            duration = parseInt64(optarg);
            break;

         case 'h':
            printHelp();
            exit(0);
            break;

         case 'r':
            should_use_recvmsg = false;
            should_use_recvfrom = true;
            break;

         case 'x':
            should_use_recvmsg = true;
            should_use_recvfrom = false;
            break;

         default:
            printf ("?? getopt returned character code 0%o ??\n", c);
         }//switch
   }//while TRUE,read all cmd-line arguments
   
   if (optind < argc) {
      printf ("non-option ARGV-elements: ");
      while (optind < argc) {
         printf ("%s ", argv[optind++]);
      }
      printf ("\n");
   }
}

//hopefully trap the TERM signal and exit gracefully
void sig_handler(int signo) { 
   //reestablish signal handler
   if ((signo != SIGILL) && (signo != SIGSEGV) && (signo != SIGBUS) && (signo != SIGFPE)) {
      signal(signo, (&sig_handler));
   }
   else {
      //cerr << "Got signal: " << signo << " setting to default handler." << endl;
      signal(signo, SIG_DFL);
   }

   if (signo == SIGTERM) {
      do_shutdown = 1;
   }
   else if (signo == SIGPIPE) {
      //ignore
   }
   else if (signo == SIGINT) {
      do_shutdown = 1;
   }
   else if (signo == SIGCHLD) {
      //Allow the system to clean up child process' resources
      // Can't log in here...races with other code it seems.
      int status = 0;
      int chld = waitpid(-1, &status, WNOHANG);
      while (chld > 0) {
         chld = waitpid(-1, &status, WNOHANG);
      }//while
   }
   else {
      cerr << "Got unknown SIGNAL:  " << signo 
             << " shutting down..." << endl;
      do_shutdown = true;
   }      
      
}//sig_handler


/** Call as:  $0 file */
int main(int argc, char** argv) {
   int r;

   signal(SIGTERM, (&sig_handler));
   signal(SIGINT, (&sig_handler));
   signal(SIGPIPE, (&sig_handler));
   signal(SIGCHLD, (&sig_handler));

   process_args(argc, argv);

   errno = 0;
   int ifi1 = if_nametoindex(dev1.c_str());
   if (ifi1 <= 0) {
      cerr << "ERROR:  Couldn't get index for dev1: " << dev1 << "  " << strerror(errno) << endl;
      printHelp();
      exit(2);
   }
   errno = 0;
   int ifi2 = if_nametoindex(dev2.c_str());
   if (ifi2 <= 0) {
      cerr << "ERROR:  Couldn't get index for dev1: " << dev2 << "  " << strerror(errno) << endl;
      printHelp();
      exit(3);
   }

   fd1 = createPacketSocket(dev1.c_str(), ETH_P_ALL, ifi1, true);
   fd2 = createPacketSocket(dev2.c_str(), ETH_P_ALL, ifi2, true);
   if (fd1 < 0 || fd2 < 0) {
      cerr << "ERROR:  Could not open packet socket on dev1: "
           << dev1 << "and/or dev2: " << dev2 << endl;
      exit(1);
   }

   fd_set input_set;
   fd_set output_set;
   fd_set exc_set;
   int maxdesc = max(fd1, fd2);
   struct timeval timeout_tv;

   uint64 started_at = getCurMs();
   uint64 now;

   while (!do_shutdown) {
      now = getCurMs();
      if (duration > 0) {
         if (started_at + duration <= now) {
            break;
         }
         else {
            uint64 togo = (started_at + duration) - now;
            timeout_tv = ms_to_tv(togo);
         }
      }
      else {
         timeout_tv.tv_sec = 4;
         timeout_tv.tv_usec = 0;
      }

      FD_ZERO(&input_set);
      FD_ZERO(&output_set);
      FD_ZERO(&exc_set);

      FD_SET(fd1, &input_set);
      FD_SET(fd2, &input_set);

      int fds = select(maxdesc + 1, &input_set, &output_set,
                       &exc_set, &timeout_tv);
      if (fds < 0) {
         if (errno != EINTR) {
            cerr << "Error in select: " << strerror(errno) << endl;
         }
      }
      else {
         if (FD_ISSET(fd1, &input_set)) {
            r = slurpDevSocket(fd1, fd2);
            rxa += r;
         }
         if (FD_ISSET(fd2, &input_set)) {
            r = slurpDevSocket(fd2, fd1);
            rxb += r;
         }
      }
   }//while true

   close(fd1);
   close(fd2);

   cout << "\n\n";
   cout << "Ran for: " << getCurMs() - started_at << "ms" << endl;
   cout << "Using rcvfrom: " << should_use_recvfrom << " rcvmsg: " << should_use_recvmsg << endl;
   cout << "rxa: " << rxa << " vlan-hdr-inserts: " << rxa_vlan_hdr_fixups 
        << " txb: " << txb << " txb_failed: " << txb_failed << endl;
   cout << "rxb: " << rxb << " vlan-hdr-inserts: " << rxb_vlan_hdr_fixups
        << " txa: " << txa << " txa_failed: " << txa_failed << endl;

   exit(0);
}//main
